@@ -70,7 +70,7 @@ RZ_API void rz_bv_fini(RZ_NONNULL RzBitVector *bv) {
  */
 RZ_API RZ_OWN RzBitVector *rz_bv_new(ut32 length) {
 	rz_return_val_if_fail(length, NULL);
-	RzBitVector *bv = RZ_NEW(RzBitVector);
+	RzBitVector *bv = RZ_NEW0(RzBitVector);
 	if (!bv || !rz_bv_init(bv, length)) {
 		free(bv);
 		return NULL;
@@ -95,7 +95,7 @@ RZ_API void rz_bv_free(RZ_NULLABLE RzBitVector *bv) {
  * \param bv RzBitVector, pointer to bitvector
  * \return str char*, bitvector string
  */
-RZ_API RZ_OWN char *rz_bv_as_string(RZ_NONNULL RzBitVector *bv) {
+RZ_API RZ_OWN char *rz_bv_as_string(RZ_NONNULL const RzBitVector *bv) {
 	rz_return_val_if_fail(bv, NULL);
 
 	char *str = (char *)malloc(bv->len + 1);
@@ -118,7 +118,7 @@ RZ_API RZ_OWN char *rz_bv_as_string(RZ_NONNULL RzBitVector *bv) {
  * \param pad whether to prepend leading zeroes to indicate the bitvector size
  * \return str char*, bitvector string in hexadecimal format
  */
-RZ_API RZ_OWN char *rz_bv_as_hex_string(RZ_NONNULL RzBitVector *bv, bool pad) {
+RZ_API RZ_OWN char *rz_bv_as_hex_string(RZ_NONNULL const RzBitVector *bv, bool pad) {
 	rz_return_val_if_fail(bv, NULL);
 
 	if (bv->len <= 64) {
@@ -177,8 +177,9 @@ RZ_API RZ_OWN RzBitVector *rz_bv_dup(const RZ_NONNULL RzBitVector *bv) {
 }
 
 /**
- * Copy from source bitvector to destination bitvector
- * the maximum copied size depends on MIN(src_len, dst_len)
+ * Copy from source bitvector to destination bitvector.
+ * The bitvectors must have the same length.
+ *
  * \param src RzBitVector, the source bitvector
  * \param dst RzBitVector, the destination bitvector
  * \return Actual size of copy
@@ -187,6 +188,7 @@ RZ_API ut32 rz_bv_copy(RZ_NONNULL const RzBitVector *src, RZ_NONNULL RzBitVector
 	rz_return_val_if_fail(src && dst, 0);
 
 	if (dst->len != src->len) {
+		rz_warn_if_reached();
 		return 0;
 	} else if (dst->len <= 64) {
 		dst->bits.small_u = src->bits.small_u;
@@ -546,6 +548,7 @@ RZ_API bool rz_bv_rshift_fill(RZ_NONNULL RzBitVector *bv, ut32 size, bool fill_b
 RZ_API RZ_OWN RzBitVector *rz_bv_and(RZ_NONNULL RzBitVector *x, RZ_NONNULL RzBitVector *y) {
 	rz_return_val_if_fail(x && y, NULL);
 	if (x->len != y->len) {
+		rz_warn_if_reached();
 		return NULL;
 	}
 
@@ -573,6 +576,7 @@ RZ_API RZ_OWN RzBitVector *rz_bv_and(RZ_NONNULL RzBitVector *x, RZ_NONNULL RzBit
 RZ_API RZ_OWN RzBitVector *rz_bv_or(RZ_NONNULL RzBitVector *x, RZ_NONNULL RzBitVector *y) {
 	rz_return_val_if_fail(x && y, NULL);
 	if (x->len != y->len) {
+		rz_warn_if_reached();
 		return NULL;
 	}
 
@@ -600,6 +604,7 @@ RZ_API RZ_OWN RzBitVector *rz_bv_or(RZ_NONNULL RzBitVector *x, RZ_NONNULL RzBitV
 RZ_API RZ_OWN RzBitVector *rz_bv_xor(RZ_NONNULL RzBitVector *x, RZ_NONNULL RzBitVector *y) {
 	rz_return_val_if_fail(x && y, NULL);
 	if (x->len != y->len) {
+		rz_warn_if_reached();
 		return NULL;
 	}
 
@@ -921,17 +926,15 @@ RZ_API RZ_OWN RzBitVector *rz_bv_sdiv(RZ_NONNULL RzBitVector *x, RZ_NONNULL RzBi
 		return ret;
 	}
 
-	if (mx && my) {
-		neg_x = rz_bv_neg(x);
-		neg_y = rz_bv_neg(y);
+	// mx && my
+	neg_x = rz_bv_neg(x);
+	neg_y = rz_bv_neg(y);
 
-		ret = rz_bv_div(neg_x, neg_y);
-		rz_bv_free(neg_x);
-		rz_bv_free(neg_y);
-		return ret;
-	}
+	ret = rz_bv_div(neg_x, neg_y);
+	rz_bv_free(neg_x);
+	rz_bv_free(neg_y);
 
-	return NULL; // something wrong
+	return ret;
 }
 
 /**
@@ -980,19 +983,16 @@ RZ_API RZ_OWN RzBitVector *rz_bv_smod(RZ_NONNULL RzBitVector *x, RZ_NONNULL RzBi
 		return ret;
 	}
 
-	if (mx && my) {
-		neg_x = rz_bv_neg(x);
-		neg_y = rz_bv_neg(y);
+	// mx && my
+	neg_x = rz_bv_neg(x);
+	neg_y = rz_bv_neg(y);
 
-		tmp = rz_bv_mod(neg_x, neg_y);
-		ret = rz_bv_neg(tmp);
-		rz_bv_free(neg_x);
-		rz_bv_free(neg_y);
-		rz_bv_free(tmp);
-		return ret;
-	}
-
-	return NULL; // something wrong
+	tmp = rz_bv_mod(neg_x, neg_y);
+	ret = rz_bv_neg(tmp);
+	rz_bv_free(neg_x);
+	rz_bv_free(neg_y);
+	rz_bv_free(tmp);
+	return ret;
 }
 
 /**
@@ -1369,7 +1369,7 @@ RZ_API void rz_bv_set_to_bytes_be(RZ_NONNULL const RzBitVector *bv, RZ_OUT RZ_NO
  * \param x BitVector
  * \return ut32 bitvector hash
  */
-ut32 rz_bv_hash(RZ_NULLABLE RzBitVector *x) {
+RZ_API ut32 rz_bv_hash(RZ_NULLABLE RzBitVector *x) {
 	ut32 h = 5381;
 	if (!x) {
 		return h;
@@ -1464,3 +1464,201 @@ RZ_API ut64 rz_bv_to_ut64(RZ_NONNULL const RzBitVector *x) {
 	}
 	return ret;
 }
+
+/**
+ * set a range of bits to bool value `b`, the range is inclusive
+ * pos_end element is also included
+ * \param bv RzBitVector
+ * \param pos_start start index of range
+ * \param pos_end end index of range
+ * \param b bool value
+ * \return return true if success, else return false
+ */
+RZ_API bool rz_bv_set_range(RZ_NONNULL RzBitVector *bv, ut32 pos_start, ut32 pos_end, bool b) {
+	rz_return_val_if_fail(bv, false);
+	if (pos_start > bv->len - 1 || pos_end > bv->len - 1) {
+		return false;
+	}
+
+	for (ut32 i = pos_start; i <= pos_end; ++i) {
+		rz_bv_set(bv, i, b);
+	}
+
+	return true;
+}
+
+/**
+ * check if bitvector's bits are all set to bit 1
+ * \param x RzBitVector
+ * \return true if all bits of bv `x` are set to 1
+ */
+RZ_API bool rz_bv_is_all_one(RZ_NONNULL const RzBitVector *x) {
+	rz_return_val_if_fail(x, false);
+	// could not use ~0 as full-vector when bits < 64
+
+	for (ut32 i = 0; i < x->len; ++i) {
+		if (rz_bv_get(x, i) == 0) {
+			return false;
+		}
+	}
+	return true;
+}
+
+/**
+ * get predecessor of bv (dec 1) in 2^n modulo
+ * \param bv
+ * \return predecessor of bv
+ */
+RZ_API RZ_OWN RzBitVector *rz_bv_pred(RZ_NONNULL RzBitVector *bv) {
+	rz_return_val_if_fail(bv, NULL);
+	ut32 len = bv->len;
+	if (len <= 64) {
+		ut64 val = rz_bv_to_ut64(bv);
+		val -= 1;
+		return rz_bv_new_from_ut64(len, val);
+	}
+
+	RzBitVector *one = rz_bv_new_one(len);
+	RzBitVector *result = rz_bv_sub(bv, one, NULL);
+	rz_bv_free(one);
+
+	return result;
+}
+
+/**
+ * get successor of bv (inc 1) in 2^n modulo
+ * \param bv
+ * \return successor of bv
+ */
+RZ_API RZ_OWN RzBitVector *rz_bv_succ(RZ_NONNULL RzBitVector *bv) {
+	rz_return_val_if_fail(bv, NULL);
+	ut32 len = bv->len;
+	if (len <= 64) {
+		ut64 val = rz_bv_to_ut64(bv);
+		val += 1;
+		return rz_bv_new_from_ut64(len, val);
+	}
+
+	RzBitVector *one = rz_bv_new_one(len);
+	RzBitVector *result = rz_bv_sub(bv, one, NULL);
+	rz_bv_free(one);
+
+	return result;
+}
+
+/**
+ * Arithmetic right shift of bv, shift right with (msb bv) bit filled
+ * \param bv
+ * \param dist shift distance
+ * \return true if success
+ */
+RZ_API bool rz_bv_arshift(RZ_NONNULL RzBitVector *bv, ut32 dist) {
+	rz_return_val_if_fail(bv, false);
+	bool msb = rz_bv_msb(bv);
+	return rz_bv_rshift_fill(bv, dist, msb);
+}
+
+/**
+ * cast bv to sort (to_size), fill with fill_bit. fill_bit has no effect if it's a narrowing cast
+ * If m = size s - size (sort b) > 0 then m bits b are prepended to the most significant part of the vector.
+ * \param bv
+ * \param to_size new bitvector length
+ * \param fill_bit specify filling bit if extend
+ * \return new bv with length (to_size)
+ */
+RZ_API RzBitVector *rz_bv_cast(RZ_NONNULL RzBitVector *bv, ut32 to_size, bool fill_bit) {
+	rz_return_val_if_fail(bv, NULL);
+
+	RzBitVector *ret = rz_bv_new(to_size);
+	rz_bv_set_all(ret, fill_bit);
+	rz_bv_copy_nbits(bv, 0, ret, 0, RZ_MIN(bv->len, to_size));
+
+	return ret;
+}
+
+/**
+ * signed cast of bv, (signed_cast x n) = (cast x n (msb x))
+ * \param bv
+ * \param to_size cast bitvector length
+ * \return new bv with length (to_size)
+ */
+RZ_API RZ_OWN RzBitVector *rz_bv_signed_cast(RZ_NONNULL RzBitVector *bv, ut32 to_size) {
+	return rz_bv_cast(bv, to_size, rz_bv_msb(bv));
+}
+
+/**
+ * unsigned cast of bv, (unsigned_cast x n) = (cast x n 0)
+ * \param bv
+ * \param to_size cast bitvector length
+ * \return new bv with length (to_size)
+ */
+RZ_API RZ_OWN RzBitVector *rz_bv_unsigned_cast(RZ_NONNULL RzBitVector *bv, ut32 to_size) {
+	return rz_bv_cast(bv, to_size, false);
+}
+
+/**
+ * strict signed less than, x < y
+ * \param x bv as signed value
+ * \param y bv as signed value
+ * \return compare result as bool value
+ */
+RZ_API bool rz_bv_slt(RZ_NONNULL RzBitVector *x, RZ_NONNULL RzBitVector *y) {
+	rz_return_val_if_fail(x && y, false);
+	// x < y === !(x >= y) === !(y <= x)
+	return !rz_bv_sle(y, x);
+}
+
+/**
+ * strict unsigned less than, x < y
+ * \param x bv as unsigned
+ * \param y bv as unsigned
+ * \return
+ */
+RZ_API bool rz_bv_ult(RZ_NONNULL RzBitVector *x, RZ_NONNULL RzBitVector *y) {
+	rz_return_val_if_fail(x && y, false);
+	return !rz_bv_ule(y, x);
+}
+
+/**
+ * strict signed great then, x > y
+ * \param x bv as signed
+ * \param y bv as signed
+ * \return
+ */
+RZ_API bool rz_bv_sgt(RZ_NONNULL RzBitVector *x, RZ_NONNULL RzBitVector *y) {
+	rz_return_val_if_fail(x && y, false);
+	return !rz_bv_sle(x, y);
+}
+
+/**
+ * strict unsigned great than, x > y
+ * \param x bv as unsigned
+ * \param y bv as unsigned
+ * \return
+ */
+RZ_API bool rz_bv_ugt(RZ_NONNULL RzBitVector *x, RZ_NONNULL RzBitVector *y) {
+	rz_return_val_if_fail(x && y, false);
+	return !rz_bv_ule(x, y);
+}
+
+/**
+ * strict signed great than or equal, x >= y
+ * \param x bv as signed
+ * \param y bv as signed
+ * \return
+ */
+RZ_API bool rz_bv_sge(RZ_NONNULL RzBitVector *x, RZ_NONNULL RzBitVector *y) {
+	rz_return_val_if_fail(x && y, false);
+	return rz_bv_sle(y, x);
+}
+
+/**
+ * strict unsigned great than or equal, x >= y
+ * \param x bv as unsigned
+ * \param y bv as unsigned
+ * \return
+ */
+RZ_API bool rz_bv_uge(RZ_NONNULL RzBitVector *x, RZ_NONNULL RzBitVector *y) {
+	rz_return_val_if_fail(x && y, false);
+	return rz_bv_ule(y, x);
+}
